@@ -1,22 +1,26 @@
 import weakref
 import asyncio
+import logging
 
 from enum import Enum
 from random import randint
-from queue import SimpleQueue, Empty
+from asyncio import TimeoutError
+from zthreading.thread_queue import Queue
+from queue import Empty
 from typing import Dict, Callable, List, Generator, AsyncGenerator
 
 
 class Event:
-    def __init__(self, name: str, args: list, kwargs: dict):
+    def __init__(self, name: str, args: list, kwargs: dict, sender=None):
         super().__init__()
         self.name = name
         self.args = args
         self.kwargs = kwargs
+        self.sender = sender
 
 
 def get_active_loop() -> asyncio.AbstractEventLoop:
-    """returns the current active asyncio loop or creates 
+    """returns the current active asyncio loop or creates
     a new one.
 
     Returns:
@@ -27,13 +31,11 @@ def get_active_loop() -> asyncio.AbstractEventLoop:
     return loop
 
 
-def filter_events_predict(name: str, args: list, kwargs: dict) -> bool:
+def filter_events_predict(event: Event) -> bool:
     """Allows or denies events from being emitted.
 
     Args:
-        name (str): The name of the event
-        args (list): The argument list
-        kwargs (dict): The named argument list.
+        event (zthreading.events.Event): The event to filer
 
     Returns:
         bool: True if the event is allowed.
@@ -46,6 +48,8 @@ class EventHandler:
     _pipeto: [] = None
     _action_last_idx = 0
     stop_all_streams_event_name = None
+    warning_event_name = "warning"
+    error_event_name = "error"
 
     def __init__(self, on_event: Callable = None):
         """An event handler. Allows,
@@ -59,7 +63,7 @@ class EventHandler:
             hndl.emit("my event", "my message)
 
         Args:
-            on_event (optional): Called on any event. Defaults to None.
+            on_event (optional, def handler(event:Event) ): Called on any event. Defaults to None.
         """
         super().__init__()
         self._pipeto = []
@@ -73,7 +77,7 @@ class EventHandler:
 
     @staticmethod
     def _get_value_from_reference(val):
-        """Returns the value of a reference or the value itself. 
+        """Returns the value of a reference or the value itself.
         Allows for easy access to weakref objects.
         """
         if isinstance(val, weakref.ReferenceType):
@@ -81,8 +85,7 @@ class EventHandler:
         return val
 
     def _create_object_instance_unique_event_name(self, base_name: str):
-        """Creates a unique name for internal event handling.
-        """
+        """Creates a unique name for internal event handling."""
         return f"{self.__class__.__name__}.{base_name} (oid: {id(self)}, rid:{randint(0,10000)})"
 
     def on_any_event(self, action: Callable) -> int:
@@ -120,7 +123,7 @@ class EventHandler:
         assert isinstance(name, str), ValueError("name must be a string")
         assert callable(action), ValueError("action must be a callable.")
 
-        if not self.hasEvent(name):
+        if not self.has_event(name):
             self._event_actions[name] = dict()
         idx = self._action_last_idx
         self._action_last_idx += 1
@@ -131,7 +134,7 @@ class EventHandler:
 
         return idx
 
-    def hasEvent(self, name: str, index: int = None):
+    def has_event(self, name: str, index: int = None):
         """Returns true if a event action is present
 
         Arguments:
@@ -168,7 +171,7 @@ class EventHandler:
 
         assert isinstance(name, str), ValueError("name must be a string")
 
-        if not self.hasEvent(name):
+        if not self.has_event(name):
             return False
 
         if idx is not None:
@@ -180,12 +183,11 @@ class EventHandler:
         return True
 
     def _get_event_actions_by_name(self, name: str) -> List[Callable]:
-        """Internal. Retruns all the events actions by name.
-        """
+        """Internal. Retruns all the events actions by name."""
         if self._event_actions_search_by_name is None or name not in self._event_actions_search_by_name:
             self._event_actions_search_by_name = self._event_actions_search_by_name or dict()
 
-            if not self.hasEvent(name):
+            if not self.has_event(name):
                 self._event_actions_search_by_name[name] = []
             else:
                 self._event_actions_search_by_name[name] = list(self._event_actions[name].values())
@@ -193,15 +195,13 @@ class EventHandler:
         return self._event_actions_search_by_name[name]
 
     def _get_catch_all_event_actions(self, original_event_name: str):
-        """Internal gets all catch all events actions by name,
-        """
-        if original_event_name != self.catch_all_event_name and self.hasEvent(self.catch_all_event_name):
+        """Internal gets all catch all events actions by name,"""
+        if original_event_name != self.catch_all_event_name and self.has_event(self.catch_all_event_name):
             return list(self._event_actions[self.catch_all_event_name].values())
         return []
 
     def _get_pipe_handlers(self) -> List["EventHandler"]:
-        """Get all pipe handlers to propagate events.
-        """
+        """Get all pipe handlers to propagate events."""
         handlers = []
         needs_pipe_cleaning = False
         for hndl in self._pipeto:
@@ -219,8 +219,7 @@ class EventHandler:
 
     @classmethod
     def _process_in_thread_event_action_result(cls, action_result, debug: bool = None):
-        """Call to await (if possible) aysncio results.
-        """
+        """Call to await (if possible) aysncio results."""
         if asyncio.iscoroutine(action_result):
             loop = get_active_loop()
             if loop.is_running():
@@ -230,8 +229,23 @@ class EventHandler:
         else:
             action_result
 
+    def _execute_event_action(self, action, event: Event, add_name: bool = False):
+        action: Callable = self._get_value_from_reference(action)
+        assert isinstance(action, Callable), "Event action is not a callable"
+        try:
+            args = event.args
+            if add_name:
+                args = [event.name, *41]
+            self._process_in_thread_event_action_result(action(*args, **event.kwargs))
+        except TypeError as err:
+            action_error = f'Failed to execute action @ File "{action.__code__.co_filename}", line {action.__code__.co_firstlineno} '
+            err.args = (*err.args, action_error)
+            raise err
+        except Exception as ex:
+            raise ex
+
     def emit(self, name: str, *args, **kwargs):
-        """Emits an event. Any arguments sent after name, will
+        """A-Synchronically emits an event. Any arguments sent after name, will
         be passed to the event action.
 
         Arguments:
@@ -241,25 +255,50 @@ class EventHandler:
             name = str(name)
         assert isinstance(name, str), ValueError("name must be a string")
 
-        if self._events_filter is not None:
-            if self._events_filter(name, args, kwargs) is not True:
-                return
+        self.emit_event(Event(name, args, kwargs, sender=self))
+
+    def emit_event(self, event: Event):
+        if self._is_event_filtered(event):
+            return
 
         if self.on_event is not None:
-            self._process_in_thread_event_action_result(self.on_event(self, name, *args, **kwargs))
+            self._process_in_thread_event_action_result(self.on_event(event))
 
-        for action in self._get_event_actions_by_name(name):
+        for action in self._get_event_actions_by_name(event.name):
             # actions are not automatically removed.
-            action = self._get_value_from_reference(action)
-            self._process_in_thread_event_action_result(action(*args, **kwargs))
+            self._execute_event_action(action, event)
 
-        for action in self._get_catch_all_event_actions(name):
+        for action in self._get_catch_all_event_actions(event.name):
             # actions are not automatically removed.
-            action = self._get_value_from_reference(action)
-            self._process_in_thread_event_action_result(action(name, *args, **kwargs))
+            self._execute_event_action(action, event, add_name=True)
 
         for handler in self._get_pipe_handlers():
-            self._process_in_thread_event_action_result(handler.emit(name, *args, **kwargs))
+            self._process_in_thread_event_action_result(handler.emit_event(event))
+
+    def bind_logger(self, logger=logging):
+        """Binds a logger to show warnings and errors. Event handler.
+
+        Args:
+            logger (logging): the logger to bind
+        """
+        self.on(self.error_event_name, lambda h, err: logger.error(err))
+        self.on(self.warning_event_name, lambda h, wrn: logger.warning(wrn))
+
+    def emit_error(self, err):
+        """Emit an error.
+
+        Args:
+            err (Exception): The error to emit
+        """
+        self.emit(self.error_event_name, self, err)
+
+    def emit_warning(self, warning):
+        """Emit a warning.
+
+        Args:
+            err (Exception): The warning to emit.
+        """
+        self.emit(self.warning_event_name, warning)
 
     def pipe(self, other: "EventHandler", use_weak_reference: bool = False):
         """Pipe all events emitted from this handler to another handler.
@@ -313,6 +352,11 @@ class EventHandler:
     def filter_events(self, predict: filter_events_predict):
         self._events_filter = predict
 
+    def _is_event_filtered(self, event: Event):
+        if self._events_filter is None:
+            return False
+        return self._events_filter(event.name, *event.args, **event.kwargs)
+
     @classmethod
     def create_events_filter_pipe(cls, predict: filter_events_predict) -> "EventHandler":
         """Faster event handler creation. The filter pipe allows for passing only specific events
@@ -353,71 +397,73 @@ class EventHandler:
 
     def stop_all_streams(self):
         """Invokes an interanl events that stops any event streams that were opened.
-        When a steam is stopped, its yield command becomes unblocking. 
-        
+        When a steam is stopped, its yield command becomes unblocking.
+
         For the case of a loop, it would exit, eg,
-        
+
         Example:
 
             for a in my_stream:
                 ....
-            
+
             hndl.stop_all_steams -> will exit loop.
         """
         self.emit(self.stop_all_streams_event_name)
 
-    def _prepare_stream_queue(self, event_name: str = None) -> (SimpleQueue, "EventHandler"):
-        """Internal, prepare a stream internal queue to manage events.
-        """
-        pipe_handler = EventHandler()
-        queue = SimpleQueue()
+    def _prepare_stream_queue(
+        self,
+        event_names: List[str] = None,
+        allow_all_events=False,
+    ) -> (Queue, "EventHandler"):
+        """Internal, prepare a stream internal queue to manage events."""
+        queue = Queue()
 
-        def append_to_queue(name, *args: list, **kwargs):
-            if event_name is not None and name not in [
-                event_name,
-                self.stop_all_streams_event_name,
-            ]:
+        allowed_events = (
+            None
+            if event_names is None
+            else set(event_names + [self.stop_all_streams_event_name, self.error_event_name])
+        )
+
+        def append_to_queue(event: Event):
+            if allowed_events is not None and event.name not in allowed_events:
                 return
-            queue.put(Event(name, args, kwargs))
+            queue.put(event)
+
+        pipe_handler = EventHandler(on_event=append_to_queue)
 
-        pipe_handler.on_any_event(append_to_queue)
         self.pipe(pipe_handler, use_weak_reference=True)
 
         return queue, pipe_handler
 
-    def _get_queue_event(self, queue: SimpleQueue, timeout: float) -> Event:
-        """Internal. Get next queued event, with timeout.
-        """
+    @classmethod
+    def _get_queue_event(cls, queue: Queue, timeout: float) -> Event:
+        """Internal. Get next queued event, with timeout."""
         try:
             event = queue.get(block=True, timeout=timeout)
         except Empty:
             raise Empty("Timed out while waiting for stream")
 
-        if event.name == "error":
-            raise event.args[0]
-
         return event
 
     def _create_stream(
         self,
-        queue: SimpleQueue,
+        queue: Queue,
         pipe_handler: "EventHander",  # noqa: F821
         timeout: float,
         process_event_data: Callable = None,
+        throw_errors: bool = True,
     ) -> Generator[Event, None, None]:
-        """Internal. Creates a new stream.
+        """Internal. Creates a new stream."""
 
-        Args:
-            queue (SimpleQueue): The associated queue.
-            pipe_handler (EventHander): The event piping handler.
-            process_event_data (Callable, optional): A method to pre-process event data before
-                the event is sent to the stream. Defaults to None.
-
-        Yields:
-            Generator of Event.
-        """
         while True:
             ev: Event = self._get_queue_event(queue, timeout)
+            if ev.name == self.error_event_name:
+                if throw_errors:
+                    raise ev.args[1] if len(ev.args) > 1 else ev.args[0] if len(ev.args) == 1 else Exception(
+                        "Unknown stream exception"
+                    )
+                else:
+                    continue
             if ev.name == self.stop_all_streams_event_name:
                 break
             if process_event_data is not None:
@@ -426,28 +472,14 @@ class EventHandler:
 
     async def _create_stream_async(
         self,
-        queue: SimpleQueue,
+        queue: Queue,
         pipe_handler: "EventHander",  # noqa: F821
         timeout: float,
         process_event_data: Callable = None,
+        throw_errors: bool = True,
     ) -> AsyncGenerator[Event, None]:
-        """Internal. Creates a new async stream.
-
-        Args:
-            queue (SimpleQueue): The associated queue.
-            pipe_handler (EventHander): The event piping handler.
-            process_event_data (Callable, optional): A method to pre-process event data before
-                the event is sent to the stream. Defaults to None.
-
-        Yields:
-            AsyncGenerator of Event.
-        """
-        while True:
-            ev: Event = self._get_queue_event(queue, timeout)
-            if ev.name == self.stop_all_streams_event_name:
-                break
-            if process_event_data is not None:
-                ev = process_event_data(ev)
+        """Internal. Creates a new async stream."""
+        for ev in self._create_stream(queue, pipe_handler, timeout, process_event_data, throw_errors):
             yield ev
 
     def stream(
@@ -456,6 +488,7 @@ class EventHandler:
         timeout: float = None,
         use_async_loop: bool = False,
         process_event_data: Callable = None,
+        throw_errors: bool = True,
     ) -> Generator[Event, None, None]:
         """Creates an event stream that will collect any event from this handler.
 
@@ -467,6 +500,7 @@ class EventHandler:
                 in an asyncio compatible stream. Defaults to False.
             process_event_data (Callable, optional): A method to be called on any event value. The
                 result of this method is passed to the stream. Defaults to None.
+            throw_errors (bool, optional): Throw error events (event_name=self.error_event_name)
 
         Yields:
             Generator|AsyncGenerator of Event | the result of process_event_data
@@ -474,14 +508,135 @@ class EventHandler:
         if isinstance(event_name, Enum):
             event_name = str(event_name)
 
-        assert event_name is None or isinstance(event_name, str), ValueError(
-            "event_name must be either a string or None"
+        if event_name is not None and not isinstance(event_name, list):
+            event_name = [event_name]
+
+        assert event_name is None or all([isinstance(v, str) for v in event_name]), ValueError(
+            "event_name must be either a string, a list of strings or None"
         )
 
         queue, pipe_handler = self._prepare_stream_queue(event_name)
         if use_async_loop is True:
-            return self._create_stream_async(queue, pipe_handler, timeout, process_event_data)
-        return self._create_stream(queue, pipe_handler, timeout, process_event_data)
+            return self._create_stream_async(
+                queue=queue,
+                pipe_handler=pipe_handler,
+                timeout=timeout,
+                process_event_data=process_event_data,
+                throw_errors=throw_errors,
+            )
+        else:
+            return self._create_stream(
+                queue=queue,
+                pipe_handler=pipe_handler,
+                timeout=timeout,
+                process_event_data=process_event_data,
+                throw_errors=throw_errors,
+            )
+
+    def wait_for(
+        self,
+        predict,
+        raise_errors: bool = True,
+        wait_count: int = 1,
+        timeout: float = None,
+    ):
+        """Waits for a specific event to be invoked.
+
+        Args:
+            predict (Callable or str): If a string, waits for the specific event. Otherwise expects true
+                when a matching event is found. (lambda task, name: ? true)
+            raise_errors (bool): True if to raise handler errors.
+            wait_count (int): How many times should the event be triggered. Must be larger then zero.
+            timeout (float): The timeout in seconds before throwing an error.
+
+        Returns:
+            The list of tasks sent to the method.
+        """
+        return self.wait_for_events(predict, [self], raise_errors=raise_errors, wait_count=wait_count, timeout=timeout)
+
+    @classmethod
+    def wait_for_events(
+        cls,
+        predict,
+        handlers: List["EventHandler"],
+        raise_errors: bool = True,
+        wait_count: int = 1,
+        timeout: float = None,
+    ) -> List["EventHandler"]:
+        """Waits for a specific event to be invoked on the handlers.
+
+        Args:
+            predict (Callable or str): If a string, waits for the specific event. Otherwise expects true
+                when a matching event is found. (lambda task, name: ? true)
+            handlers (list) : List of handlers to execute for.
+            raise_errors (bool): True if to raise handler errors.
+            wait_count (int): How many times should the event be triggered. Must be larger then zero.
+            timeout (float): The timeout in seconds before throwing an error.
+
+        Returns:
+            The list of tasks sent to the method.
+        """
+
+        if isinstance(handlers, EventHandler):
+            handlers: List["Task"] = [handlers]
+
+        assert wait_count > 0, "Wait count must be at least 1"
+
+        if predict is not None and not isinstance(predict, Callable):
+            predict_equals = predict
+
+            def predict_event_by_name(sender: EventHandler, event: Event):
+                return event.name == predict_equals
+
+            predict = predict_event_by_name
+
+        assert predict is None or callable(predict), "Predict must be a Callable or event name string"
+
+        wait_queue = Queue()
+        first_error = None
+        matched_handlers = []
+
+        def stop_on_error(hndl: EventHandler, error):
+            nonlocal first_error
+            first_error = error
+            wait_queue.put(False)
+
+        def on_piped_event(handler: EventHandler, event: Event):
+            if event.name == handler.error_event_name:
+                stop_on_error(handler, event.args[1])
+                return
+            try:
+                if predict is None or predict(handler, event):
+                    matched_handlers.append(handler)
+                if len(matched_handlers) == wait_count:
+                    wait_queue.put(True)
+            except Exception as err:
+                stop_on_error(handler, err)
+
+        pipes = []
+
+        def bind(handler: EventHandler):
+            pipe_handler = EventHandler(on_event=lambda event: on_piped_event(handler, event))
+            pipes.append(pipe_handler)
+            handler.pipe(
+                pipe_handler,
+                use_weak_reference=True,
+            )
+
+        for handler in handlers:
+            bind(handler)
+
+        try:
+            wait_queue.get(timeout=timeout)
+        except Empty:
+            first_error = TimeoutError(f"Timeout while waiting for event: {predict}")
+
+        if raise_errors and first_error is not None:
+            raise first_error
+        elif first_error is not None:
+            return first_error
+
+        return matched_handlers
 
 
 class AsyncEventHandler(EventHandler):
@@ -503,13 +658,27 @@ class AsyncEventHandler(EventHandler):
 
     @classmethod
     async def _process_async_action_result(cls, action_result):
-        """Call to await (if needed) aysncio results.
-        """
+        """Call to await (if needed) aysncio results."""
         if asyncio.iscoroutine(action_result):
             return await action_result
         else:
             return action_result
 
+    async def _execute_event_action(self, action, event: Event, add_name: bool = False):
+        action: Callable = self._get_value_from_reference(action)
+        assert isinstance(action, Callable), "Event action is not a callable"
+        try:
+            args = event.args
+            if add_name:
+                args = [event.name, *args]
+            await self._process_async_action_result(action(*args, **event.kwargs))
+        except TypeError as err:
+            action_error = f'Failed to execute action @ File "{action.__code__.co_filename}", line {action.__code__.co_firstlineno} '
+            err.args = (*err.args, action_error)
+            raise err
+        except Exception as ex:
+            raise ex
+
     async def emit(self, name: str, *args, **kwargs):
         """A-Synchronically emits an event. Any arguments sent after name, will
         be passed to the event action.
@@ -521,21 +690,25 @@ class AsyncEventHandler(EventHandler):
             name = str(name)
         assert isinstance(name, str), ValueError("name must be a string")
 
+        await self.emit_event(Event(name, args, kwargs, sender=self))
+
+    async def emit_event(self, event: Event):
+        if self._is_event_filtered(event):
+            return
+
         if self.on_event is not None:
-            await self._process_async_action_result(self.on_event(name, *args, **kwargs))
+            await self._process_async_action_result(self.on_event(self, event.name, *event.args, **event.kwargs))
 
-        for action in self._get_event_actions_by_name(name):
+        for action in self._get_event_actions_by_name(event.name):
             # actions are not automatically removed.
-            action = self._get_value_from_reference(action)
-            await self._process_async_action_result(action(*args, **kwargs))
+            await self._execute_event_action(action, event)
 
-        for action in self._get_catch_all_event_actions(name):
+        for action in self._get_catch_all_event_actions(event.name):
             # actions are not automatically removed.
-            action = self._get_value_from_reference(action)
-            await self._process_async_action_result(action(name, *args, **kwargs))
+            await self._execute_event_action(action, event, add_name=True)
 
         for handler in self._get_pipe_handlers():
-            await self._process_async_action_result(handler.emit(name, *args, **kwargs))
+            await self._process_async_action_result(handler.emit_event(event))
 
     def emit_sync(self, name: str, *args, **kwargs):
         """Synchronically emits an event. Any arguments sent after name, will
